@@ -41,18 +41,14 @@
         "milter", priv->pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_milter(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "milter", priv->pool->tag.uid, \
+#define msg_debug_milter(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_milter_log_id, "milter", priv->pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
 
-struct rspamd_milter_context {
-	gchar *spam_header;
-	void *sessions_cache;
-	gboolean discard_on_reject;
-};
+INIT_LOG_MODULE(milter)
 
-static struct rspamd_milter_context *milter_ctx = NULL;
+static const struct rspamd_milter_context *milter_ctx = NULL;
 
 static gboolean  rspamd_milter_handle_session (
 		struct rspamd_milter_session *session,
@@ -99,6 +95,8 @@ rspamd_milter_session_reset (struct rspamd_milter_session *session,
 	guint i;
 
 	if (how & RSPAMD_MILTER_RESET_IO) {
+		msg_debug_milter ("cleanup IO on abort");
+
 		DL_FOREACH_SAFE (priv->out_chain, obuf, obuf_tmp) {
 			rspamd_milter_obuf_free (obuf);
 		}
@@ -111,46 +109,57 @@ rspamd_milter_session_reset (struct rspamd_milter_session *session,
 	}
 
 	if (how & RSPAMD_MILTER_RESET_COMMON) {
+		msg_debug_milter ("cleanup common data on abort");
+
 		if (session->message) {
 			session->message->len = 0;
+			msg_debug_milter ("cleanup message on abort");
 		}
 
 		if (session->rcpts) {
 			PTR_ARRAY_FOREACH (session->rcpts, i, cur) {
-				rspamd_email_address_unref (cur);
+				rspamd_email_address_free (cur);
 			}
 
+			msg_debug_milter ("cleanup %d recipients on abort",
+					(gint)session->rcpts->len);
+
 			g_ptr_array_free (session->rcpts, TRUE);
 			session->rcpts = NULL;
 		}
 
 		if (session->from) {
-			rspamd_email_address_unref (session->from);
+			msg_debug_milter ("cleanup from");
+			rspamd_email_address_free (session->from);
 			session->from = NULL;
 		}
 
 		if (session->helo) {
+			msg_debug_milter ("cleanup helo");
 			session->helo->len = 0;
 		}
 
-		if (session->hostname) {
-			session->hostname->len = 0;
-		}
-
 		if (priv->headers) {
+			msg_debug_milter ("cleanup headers");
 			g_hash_table_remove_all (priv->headers);
 		}
 	}
 
 	if (how & RSPAMD_MILTER_RESET_ADDR) {
 		if (session->addr) {
+			msg_debug_milter ("cleanup addr");
 			rspamd_inet_address_free (session->addr);
 			session->addr = NULL;
 		}
+		if (session->hostname) {
+			msg_debug_milter ("cleanup hostname");
+			session->hostname->len = 0;
+		}
 	}
 
 	if (how & RSPAMD_MILTER_RESET_MACRO) {
 		if (session->macros) {
+			msg_debug_milter ("cleanup macros");
 			g_hash_table_unref (session->macros);
 			session->macros = NULL;
 		}
@@ -311,6 +320,8 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 		else {
 			guchar proto;
 			guint16 port;
+			gchar ip6_str[INET6_ADDRSTRLEN + 3];
+			gsize r;
 
 			/*
 			 * Important notice: Postfix do NOT use this command to pass
@@ -319,10 +330,14 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 			 */
 			if (session->hostname == NULL) {
 				session->hostname = rspamd_fstring_new_init (pos, zero - pos);
+				msg_debug_milter ("got hostname on connect phase: %V",
+						session->hostname);
 			}
 			else {
 				session->hostname = rspamd_fstring_assign (session->hostname,
 						pos, zero - pos);
+				msg_debug_milter ("rewrote hostname on connect phase: %V",
+						session->hostname);
 			}
 
 			pos = zero + 1;
@@ -373,16 +388,28 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 						break;
 
 					case RSPAMD_MILTER_CONN_INET6:
-						session->addr = rspamd_inet_address_new (AF_INET, NULL);
+						session->addr = rspamd_inet_address_new (AF_INET6, NULL);
 
 						if (zero - pos > sizeof ("IPv6:") &&
 								rspamd_lc_cmp (pos, "IPv6:",
 										sizeof ("IPv6:") - 1) == 0) {
 							/* Kill sendmail please */
 							pos += sizeof ("IPv6:") - 1;
+
+							if (*pos != '[') {
+								/* Add explicit braces */
+								r = rspamd_snprintf (ip6_str, sizeof (ip6_str),
+										"[%*s]", (int)(zero - pos), pos);
+							}
+							else {
+								r = rspamd_strlcpy (ip6_str, pos, sizeof (ip6_str));
+							}
+						}
+						else {
+							r = rspamd_strlcpy (ip6_str, pos, sizeof (ip6_str));
 						}
 
-						if (!rspamd_parse_inet_address_ip (pos, zero - pos,
+						if (!rspamd_parse_inet_address_ip (ip6_str, r,
 								session->addr)) {
 							err = g_error_new (rspamd_milter_quark (), EINVAL,
 									"invalid connect command (bad IPv6)");
@@ -433,7 +460,7 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 		while (pos < end) {
 			zero = memchr (pos, '\0', cmdlen);
 
-			if (zero == NULL) {
+			if (zero == NULL || zero >= end) {
 				err = g_error_new (rspamd_milter_quark (), EINVAL, "invalid "
 						"macro command (no name)");
 				rspamd_milter_on_protocol_error (session, priv, err);
@@ -445,9 +472,9 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 				rspamd_ftok_t *name_tok, *value_tok;
 				const guchar *zero_val;
 
-				zero_val = memchr (zero + 1, '\0', cmdlen);
+				zero_val = memchr (zero + 1, '\0',  end - zero - 1);
 
-				if (end > zero_val) {
+				if (zero_val != NULL && end > zero_val) {
 					name = rspamd_fstring_new_init (pos, zero - pos);
 					value = rspamd_fstring_new_init (zero + 1,
 							zero_val - zero - 1);
@@ -458,11 +485,6 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 					msg_debug_milter ("got macro: %T -> %T",
 							name_tok, value_tok);
 
-					if (rspamd_ftok_cstr_equal (name_tok, "{mail_host}", FALSE)) {
-						session->hostname = rspamd_fstring_assign (session->hostname,
-								value_tok->begin, value_tok->len);
-					}
-
 					cmdlen -= zero_val - pos;
 					pos = zero_val + 1;
 				}
@@ -515,6 +537,10 @@ rspamd_milter_process_command (struct rspamd_milter_session *session,
 		break;
 	case RSPAMD_MILTER_CMD_HEADER:
 		msg_debug_milter ("got header command");
+		if (!session->message) {
+			session->message = rspamd_fstring_sized_new (
+					RSPAMD_MILTER_MESSAGE_CHUNK);
+		}
 		zero = memchr (pos, '\0', cmdlen);
 
 		if (zero == NULL) {
@@ -764,7 +790,10 @@ rspamd_milter_consume_input (struct rspamd_milter_session *session,
 	end = priv->parser.buf->str + priv->parser.buf->len;
 
 	while (p < end) {
-		msg_debug_milter("offset: %d, state: %d", (gint)(p - (const guchar *)priv->parser.buf->str), priv->parser.state);
+		msg_debug_milter("offset: %d, state: %d",
+				(gint)(p - (const guchar *)priv->parser.buf->str),
+				priv->parser.state);
+
 		switch (priv->parser.state) {
 		case st_len_1:
 			/* The first length byte in big endian order */
@@ -882,6 +911,7 @@ rspamd_milter_consume_input (struct rspamd_milter_session *session,
 	if (p == end) {
 		priv->parser.buf->len = 0;
 		priv->parser.pos = 0;
+		priv->parser.cmd_start = 0;
 	}
 
 	if (priv->out_chain) {
@@ -945,6 +975,8 @@ rspamd_milter_handle_session (struct rspamd_milter_session *session,
 
 			return rspamd_milter_consume_input (session, priv);
 		}
+
+		break;
 	case RSPAMD_MILTER_WRITE_REPLY:
 	case RSPAMD_MILTER_WRITE_AND_DIE:
 		if (priv->out_chain == NULL) {
@@ -986,6 +1018,8 @@ rspamd_milter_handle_session (struct rspamd_milter_session *session,
 						priv->err_cb (priv->fd, session, priv->ud, err);
 						REF_RELEASE (session);
 						g_error_free (err);
+
+						return FALSE;
 					}
 				}
 				else if (r == 0) {
@@ -995,6 +1029,8 @@ rspamd_milter_handle_session (struct rspamd_milter_session *session,
 					priv->err_cb (priv->fd, session, priv->ud, err);
 					REF_RELEASE (session);
 					g_error_free (err);
+
+					return FALSE;
 				}
 				else {
 					if (r == to_write) {
@@ -1051,6 +1087,8 @@ rspamd_milter_handle_socket (gint fd, const struct timeval *tv,
 	priv->ev_base = ev_base;
 	priv->state = RSPAMD_MILTER_READ_MORE;
 	priv->pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), "milter");
+	priv->discard_on_reject = milter_ctx->discard_on_reject;
+	priv->quarantine_on_reject = milter_ctx->quarantine_on_reject;
 
 	if (pool) {
 		/* Copy tag */
@@ -1116,7 +1154,9 @@ rspamd_milter_send_action (struct rspamd_milter_session *session,
 	va_list ap;
 	guchar cmd, *pos;
 	rspamd_fstring_t *reply = NULL;
+	gsize len;
 	GString *name, *value;
+	const char *reason;
 	struct rspamd_milter_outbuf *obuf;
 	struct rspamd_milter_private *priv = session->priv;
 
@@ -1133,6 +1173,17 @@ rspamd_milter_send_action (struct rspamd_milter_session *session,
 		/* No additional arguments */
 		SET_COMMAND (cmd, 0, reply, pos);
 		break;
+	case RSPAMD_MILTER_QUARANTINE:
+		reason = va_arg (ap, const char *);
+
+		if (reason == NULL) {
+			reason = "";
+		}
+
+		len = strlen (reason);
+		SET_COMMAND (cmd, len + 1, reply, pos);
+		memcpy (pos, reason, len + 1);
+		break;
 	case RSPAMD_MILTER_ADDHEADER:
 		name = va_arg (ap, GString *);
 		value = va_arg (ap, GString *);
@@ -1144,6 +1195,7 @@ rspamd_milter_send_action (struct rspamd_milter_session *session,
 		memcpy (pos, value->str, value->len + 1);
 		break;
 	case RSPAMD_MILTER_CHGHEADER:
+	case RSPAMD_MILTER_INSHEADER:
 		idx = htonl (va_arg (ap, guint32));
 		name = va_arg (ap, GString *);
 		value = va_arg (ap, GString *);
@@ -1160,6 +1212,7 @@ rspamd_milter_send_action (struct rspamd_milter_session *session,
 	case RSPAMD_MILTER_REPLYCODE:
 	case RSPAMD_MILTER_ADDRCPT:
 	case RSPAMD_MILTER_DELRCPT:
+	case RSPAMD_MILTER_CHGFROM:
 		/* Single GString * argument */
 		value = va_arg (ap, GString *);
 		SET_COMMAND (cmd, value->len + 1, reply, pos);
@@ -1243,6 +1296,7 @@ rspamd_milter_macro_http (struct rspamd_milter_session *session,
 		struct rspamd_http_message *msg)
 {
 	rspamd_ftok_t *found, srch;
+	struct rspamd_milter_private *priv = session->priv;
 
 	/*
 	 * We assume postfix macros here, sendmail ones might be slightly
@@ -1295,10 +1349,50 @@ rspamd_milter_macro_http (struct rspamd_milter_session *session,
 				found->begin, found->len);
 	}
 
+	IF_MACRO("{rcpt_mailer}") {
+		rspamd_http_message_add_header_len (msg, MAILER_HEADER,
+				found->begin, found->len);
+	}
+
+	if (milter_ctx->client_ca_name) {
+		IF_MACRO ("{cert_issuer}") {
+			rspamd_http_message_add_header_len (msg, CERT_ISSUER_HEADER,
+					found->begin, found->len);
+
+			if (found->len == strlen (milter_ctx->client_ca_name) &&
+					rspamd_cryptobox_memcmp (found->begin,
+							milter_ctx->client_ca_name, found->len) == 0) {
+				msg_debug_milter ("process certificate issued by %T", found);
+				IF_MACRO("{cert_subject}") {
+					rspamd_http_message_add_header_len (msg, USER_HEADER,
+							found->begin, found->len);
+				}
+			}
+			else {
+				msg_debug_milter ("skip certificate issued by %T", found);
+			}
+
+
+		}
+	}
+	else {
+		IF_MACRO ("{cert_issuer}") {
+			rspamd_http_message_add_header_len (msg, CERT_ISSUER_HEADER,
+					found->begin, found->len);
+		}
+	}
+
 	if (!session->hostname || session->hostname->len == 0) {
 		IF_MACRO("{client_name}") {
-			rspamd_http_message_add_header_len (msg, HOSTNAME_HEADER,
-					found->begin, found->len);
+			if (!(found->len == sizeof ("unknown") - 1 &&
+					memcmp (found->begin, "unknown",
+							sizeof ("unknown") - 1) == 0)) {
+				rspamd_http_message_add_header_len (msg, HOSTNAME_HEADER,
+						found->begin, found->len);
+			}
+			else {
+				msg_debug_milter ("skip unknown hostname from being added");
+			}
 		}
 	}
 
@@ -1328,6 +1422,7 @@ rspamd_milter_to_http (struct rspamd_milter_session *session)
 	struct rspamd_http_message *msg;
 	guint i;
 	struct rspamd_email_address *rcpt;
+	struct rspamd_milter_private *priv = session->priv;
 
 	g_assert (session != NULL);
 
@@ -1341,9 +1436,16 @@ rspamd_milter_to_http (struct rspamd_milter_session *session)
 		session->message = NULL;
 	}
 
-	if (session->hostname && session->hostname->len > 0) {
-		rspamd_http_message_add_header_fstr (msg, HOSTNAME_HEADER,
-				session->hostname);
+	if (session->hostname && RSPAMD_FSTRING_LEN (session->hostname) > 0) {
+		if (!(session->hostname->len == sizeof ("unknown") - 1 &&
+				memcmp (RSPAMD_FSTRING_DATA (session->hostname), "unknown",
+						sizeof ("unknown") - 1) == 0)) {
+			rspamd_http_message_add_header_fstr (msg, HOSTNAME_HEADER,
+					session->hostname);
+		}
+		else {
+			msg_debug_milter ("skip unknown hostname from being added");
+		}
 	}
 
 	if (session->helo && session->helo->len > 0) {
@@ -1364,8 +1466,14 @@ rspamd_milter_to_http (struct rspamd_milter_session *session)
 	}
 
 	if (session->addr) {
-		rspamd_http_message_add_header (msg, IP_ADDR_HEADER,
-				rspamd_inet_address_to_string (session->addr));
+		if (rspamd_inet_address_get_af (session->addr) != AF_UNIX) {
+			rspamd_http_message_add_header (msg, IP_ADDR_HEADER,
+					rspamd_inet_address_to_string_pretty (session->addr));
+		}
+		else {
+			rspamd_http_message_add_header (msg, IP_ADDR_HEADER,
+					rspamd_inet_address_to_string (session->addr));
+		}
 	}
 
 	rspamd_milter_macro_http (session, msg);
@@ -1440,7 +1548,9 @@ rspamd_milter_process_milter_block (struct rspamd_milter_session *session,
 {
 	const ucl_object_t *elt, *cur, *cur_elt;
 	ucl_object_iter_t it;
+	struct rspamd_milter_private *priv = session->priv;
 	GString *hname, *hvalue;
+	gint idx = -1;
 
 	if (obj && ucl_object_type (obj) == UCL_OBJECT) {
 		elt = ucl_object_lookup (obj, "remove_headers");
@@ -1464,12 +1574,18 @@ rspamd_milter_process_milter_block (struct rspamd_milter_session *session,
 		/*
 		 * add_headers: {"name": "value", ... }
 		 * name could have multiple values
+		 * -or- (since 1.7)
+		 * {"name": {"value": "val", "order": 0}, ... }
 		 */
 		if (elt && ucl_object_type (elt) == UCL_OBJECT) {
 			it = NULL;
 
 			while ((cur = ucl_object_iterate (elt, &it, true)) != NULL) {
-				LL_FOREACH (cur, cur_elt) {
+				ucl_object_iter_t *elt_it;
+
+				elt_it = ucl_object_iterate_new (cur);
+
+				while ((cur_elt = ucl_object_iterate_safe (elt_it, true)) != NULL) {
 					if (ucl_object_type (cur_elt) == UCL_STRING) {
 						hname = g_string_new (ucl_object_key (cur));
 						hvalue = g_string_new (ucl_object_tostring (cur_elt));
@@ -1480,9 +1596,77 @@ rspamd_milter_process_milter_block (struct rspamd_milter_session *session,
 						g_string_free (hname, TRUE);
 						g_string_free (hvalue, TRUE);
 					}
+					else if (ucl_object_type (cur_elt) == UCL_OBJECT) {
+						const ucl_object_t *val;
+
+						val = ucl_object_lookup (cur_elt, "value");
+
+						if (val && ucl_object_type (val) == UCL_STRING) {
+							const ucl_object_t *idx_obj;
+
+							idx_obj = ucl_object_lookup_any (cur_elt, "order",
+									"index", NULL);
+							if (idx_obj) {
+								idx = ucl_object_toint (idx_obj);
+							}
+
+							hname = g_string_new (ucl_object_key (cur));
+							hvalue = g_string_new (ucl_object_tostring (val));
+
+							if (idx >= 0) {
+								rspamd_milter_send_action (session,
+										RSPAMD_MILTER_INSHEADER,
+										idx,
+										hname, hvalue);
+							}
+							else {
+								rspamd_milter_send_action (session,
+										RSPAMD_MILTER_ADDHEADER,
+										hname, hvalue);
+							}
+
+							g_string_free (hname, TRUE);
+							g_string_free (hvalue, TRUE);
+						}
+					}
 				}
+
+				ucl_object_iterate_free (elt_it);
 			}
 		}
+
+		elt = ucl_object_lookup (obj, "change_from");
+
+		if (elt && ucl_object_type (elt) == UCL_STRING) {
+			hvalue = g_string_new (ucl_object_tostring (elt));
+			rspamd_milter_send_action (session,
+					RSPAMD_MILTER_CHGFROM,
+					hvalue);
+			g_string_free (hvalue, TRUE);
+		}
+
+		elt = ucl_object_lookup (obj, "reject");
+
+		if (elt && ucl_object_type (elt) == UCL_STRING) {
+			if (strcmp (ucl_object_tostring (elt), "discard") == 0) {
+				priv->discard_on_reject = TRUE;
+				msg_info_milter ("discard message instead of rejection");
+			}
+			else if (strcmp (ucl_object_tostring (elt), "quarantine") == 0) {
+				priv->quarantine_on_reject = TRUE;
+				msg_info_milter ("quarantine message instead of rejection");
+			}
+			else {
+				priv->discard_on_reject = FALSE;
+				priv->quarantine_on_reject = FALSE;
+			}
+		}
+
+		elt = ucl_object_lookup (obj, "no_action");
+
+		if (elt && ucl_object_type (elt) == UCL_BOOLEAN) {
+			priv->no_action = ucl_object_toboolean (elt);
+		}
 	}
 
 	if (action == METRIC_ACTION_ADD_HEADER) {
@@ -1536,6 +1720,7 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 {
 	const ucl_object_t *elt;
 	struct rspamd_milter_private *priv = session->priv;
+	const gchar *str_action;
 	gint action = METRIC_ACTION_REJECT;
 	rspamd_fstring_t *xcode = NULL, *rcode = NULL, *reply = NULL;
 	GString *hname, *hvalue;
@@ -1545,7 +1730,7 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 		msg_err_milter ("cannot find scan results, tempfail");
 		rspamd_milter_send_action (session, RSPAMD_MILTER_TEMPFAIL);
 
-		return;
+		goto cleanup;
 	}
 
 	elt = ucl_object_lookup (results, "action");
@@ -1554,10 +1739,11 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 		msg_err_milter ("cannot find action in results, tempfail");
 		rspamd_milter_send_action (session, RSPAMD_MILTER_TEMPFAIL);
 
-		return;
+		goto cleanup;
 	}
 
-	rspamd_action_from_str (ucl_object_tostring (elt), &action);
+	str_action = ucl_object_tostring (elt);
+	rspamd_action_from_str (str_action, &action);
 
 	elt = ucl_object_lookup (results, "messages");
 	if (elt) {
@@ -1587,8 +1773,8 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 		hname = g_string_new (RSPAMD_MILTER_DKIM_HEADER);
 		hvalue = g_string_new (ucl_object_tostring (elt));
 
-		rspamd_milter_send_action (session, RSPAMD_MILTER_ADDHEADER,
-				hname, hvalue);
+		rspamd_milter_send_action (session, RSPAMD_MILTER_INSHEADER,
+				1, hname, hvalue);
 		g_string_free (hname, TRUE);
 		g_string_free (hvalue, TRUE);
 	}
@@ -1597,11 +1783,31 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 		goto cleanup;
 	}
 
+	if (priv->no_action) {
+		msg_info_milter ("do not apply action %s, no_action is set",
+				str_action);
+		hname = g_string_new (RSPAMD_MILTER_ACTION_HEADER);
+		hvalue = g_string_new (str_action);
+
+		rspamd_milter_send_action (session, RSPAMD_MILTER_ADDHEADER,
+				hname, hvalue);
+		g_string_free (hname, TRUE);
+		g_string_free (hvalue, TRUE);
+		rspamd_milter_send_action (session, RSPAMD_MILTER_ACCEPT);
+
+		goto cleanup;
+	}
+
 	switch (action) {
 	case METRIC_ACTION_REJECT:
-		if (milter_ctx->discard_on_reject) {
+		if (priv->discard_on_reject) {
 			rspamd_milter_send_action (session, RSPAMD_MILTER_DISCARD);
 		}
+		else if (priv->quarantine_on_reject) {
+			/* TODO: be more flexible about SMTP messages */
+			rspamd_milter_send_action (session, RSPAMD_MILTER_QUARANTINE,
+					RSPAMD_MILTER_QUARANTINE_MESSAGE);
+		}
 		else {
 			rcode = rspamd_fstring_new_init (RSPAMD_MILTER_RCODE_REJECT,
 					sizeof (RSPAMD_MILTER_RCODE_REJECT) - 1);
@@ -1609,12 +1815,18 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 					sizeof (RSPAMD_MILTER_XCODE_REJECT) - 1);
 
 			if (!reply) {
-				reply = rspamd_fstring_new_init (RSPAMD_MILTER_REJECT_MESSAGE,
-						sizeof (RSPAMD_MILTER_REJECT_MESSAGE) - 1);
+				if (milter_ctx->reject_message == NULL) {
+					reply = rspamd_fstring_new_init (
+							RSPAMD_MILTER_REJECT_MESSAGE,
+							sizeof (RSPAMD_MILTER_REJECT_MESSAGE) - 1);
+				}
+				else {
+					reply = rspamd_fstring_new_init (milter_ctx->reject_message,
+							strlen (milter_ctx->reject_message));
+				}
 			}
 
 			rspamd_milter_set_reply (session, rcode, xcode, reply);
-			rspamd_milter_send_action (session, RSPAMD_MILTER_REJECT);
 		}
 		break;
 	case METRIC_ACTION_SOFT_REJECT:
@@ -1629,7 +1841,6 @@ rspamd_milter_send_task_results (struct rspamd_milter_session *session,
 		}
 
 		rspamd_milter_set_reply (session, rcode, xcode, reply);
-		rspamd_milter_send_action (session, RSPAMD_MILTER_REJECT);
 		break;
 
 	case METRIC_ACTION_REWRITE_SUBJECT:
@@ -1674,28 +1885,14 @@ cleanup:
 	rspamd_fstring_free (rcode);
 	rspamd_fstring_free (xcode);
 	rspamd_fstring_free (reply);
+
+	rspamd_milter_session_reset (session, RSPAMD_MILTER_RESET_ABORT);
 }
 
 void
-rspamd_milter_init_library (const gchar *spam_header, void *sessions_cache,
-		gboolean discard_on_reject)
+rspamd_milter_init_library (const struct rspamd_milter_context *ctx)
 {
-	if (milter_ctx) {
-		g_free (milter_ctx->spam_header);
-		g_free (milter_ctx);
-	}
-
-	milter_ctx = g_malloc (sizeof (*milter_ctx));
-
-	if (spam_header) {
-		milter_ctx->spam_header = g_strdup (spam_header);
-	}
-	else {
-		milter_ctx->spam_header = g_strdup (RSPAMD_MILTER_SPAM_HEADER);
-	}
-
-	milter_ctx->sessions_cache = sessions_cache;
-	milter_ctx->discard_on_reject = discard_on_reject;
+	milter_ctx = ctx;
 }
 
 rspamd_mempool_t *
