@@ -198,7 +198,7 @@ rspamd_tokenizer_get_word (rspamd_stat_token_t * buf,
 	p = s;
 	token->begin = s;
 
-	for (i = 0; i <= remain; ) {
+	for (i = 0; i < remain; ) {
 		p = &s[i];
 		U8_NEXT (s, i, remain, uc); /* This also advances i */
 
@@ -254,6 +254,14 @@ rspamd_tokenizer_get_word (rspamd_stat_token_t * buf,
 		}
 	}
 
+	/* Last character */
+	if (state == feed_token) {
+		p = &s[i];
+		goto set_token;
+	}
+
+	return FALSE;
+
 set_token:
 	if (rl) {
 		*rl = processed;
@@ -330,7 +338,7 @@ rspamd_tokenize_text (gchar *text, gsize len, gboolean is_utf,
 {
 	rspamd_stat_token_t token, buf;
 	const gchar *pos = NULL;
-	gsize l;
+	gsize l = 0;
 	GArray *res;
 	GList *cur = exceptions;
 	token_get_function func;
