@@ -24,7 +24,7 @@
 #include "ref.h"
 
 typedef void (*rspamd_map_dtor) (gpointer p);
-
+extern guint rspamd_map_log_id;
 #define msg_err_map(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
 		"map", map->tag, \
         G_STRFUNC, \
@@ -37,8 +37,8 @@ typedef void (*rspamd_map_dtor) (gpointer p);
 		"map", map->tag, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_map(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "map", map->tag, \
+#define msg_debug_map(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_map_log_id, "map", map->tag, \
         G_STRFUNC, \
         __VA_ARGS__)
 
@@ -49,17 +49,48 @@ enum fetch_proto {
 	MAP_PROTO_STATIC
 };
 
+/**
+ * Data specific to file maps
+ */
+struct file_map_data {
+	gchar *filename;
+	struct stat st;
+};
+
+/**
+ * Data specific to HTTP maps
+ */
+struct http_map_data {
+	gchar *path;
+	gchar *host;
+	gchar *last_signature;
+	rspamd_fstring_t *etag;
+	time_t last_modified;
+	time_t last_checked;
+	gboolean request_sent;
+	guint64 gen;
+	guint16 port;
+};
+
+struct static_map_data {
+	guchar *data;
+	gsize len;
+	gboolean processed;
+};
+
+union rspamd_map_backend_data {
+	struct file_map_data *fd;
+	struct http_map_data *hd;
+	struct static_map_data *sd;
+};
+
 struct rspamd_map_backend {
 	enum fetch_proto protocol;
 	gboolean is_signed;
 	gboolean is_compressed;
 	guint32 id;
 	struct rspamd_cryptobox_pubkey *trusted_pubkey;
-	union {
-		struct file_map_data *fd;
-		struct http_map_data *hd;
-		struct static_map_data *sd;
-	} data;
+	union rspamd_map_backend_data data;
 	gchar *uri;
 	ref_entry_t ref;
 };
@@ -82,46 +113,23 @@ struct rspamd_map {
 	gchar *description;
 	gchar *name;
 	guint32 id;
+	gboolean scheduled_check;
+	rspamd_map_dtor dtor;
+	gpointer dtor_data;
+	rspamd_map_traverse_function traverse_function;
+	gpointer lua_map;
+	gsize nelts;
+	guint64 digest;
 	/* Should we check HTTP or just load cached data */
-	gboolean active_http;
 	struct timeval tv;
 	gdouble poll_timeout;
 	time_t next_check;
+	gboolean active_http;
 	/* Shared lock for temporary disabling of map reading (e.g. when this map is written by UI) */
 	gint *locked;
 	/* Shared cache data */
 	struct rspamd_map_cachepoint *cache;
 	gchar tag[MEMPOOL_UID_LEN];
-	rspamd_map_dtor dtor;
-	gpointer dtor_data;
-};
-
-/**
- * Data specific to file maps
- */
-struct file_map_data {
-	gchar *filename;
-	struct stat st;
-};
-
-/**
- * Data specific to HTTP maps
- */
-struct http_map_data {
-	gchar *path;
-	gchar *host;
-	gchar *last_signature;
-	time_t last_modified;
-	time_t last_checked;
-	gboolean request_sent;
-	guint64 gen;
-	guint16 port;
-};
-
-struct static_map_data {
-	guchar *data;
-	gsize len;
-	gboolean processed;
 };
 
 enum rspamd_map_http_stage {
@@ -139,8 +147,8 @@ struct map_periodic_cbdata {
 	struct event ev;
 	gboolean need_modify;
 	gboolean errored;
-	guint cur_backend;
 	gboolean locked;
+	guint cur_backend;
 	ref_entry_t ref;
 };
 
