@@ -18,7 +18,13 @@
 #include "cryptobox.h"
 #include "url.h"
 #include "str_util.h"
+#include "logger.h"
 #include "contrib/t1ha/t1ha.h"
+#include <unicode/uversion.h>
+#include <unicode/ucnv.h>
+#if U_ICU_VERSION_MAJOR_NUM >= 44
+#include <unicode/unorm2.h>
+#endif
 #include <math.h>
 
 const guchar lc_map[256] = {
@@ -79,10 +85,12 @@ rspamd_str_lc (gchar *str, guint size)
 	switch (leftover) {
 	case 3:
 		*dest++ = lc_map[(guchar)str[i++]];
+		/* FALLTHRU */
 	case 2:
 		*dest++ = lc_map[(guchar)str[i++]];
+		/* FALLTHRU */
 	case 1:
-		*dest++ = lc_map[(guchar)str[i]];
+		*dest = lc_map[(guchar)str[i]];
 	}
 
 }
@@ -120,13 +128,12 @@ rspamd_lc_cmp (const gchar *s, const gchar *d, gsize l)
 	}
 
 	while (leftover > 0) {
-		if (g_ascii_tolower (*s) != g_ascii_tolower (*d)) {
-			return (*s) - (*d);
+		if (g_ascii_tolower (s[i]) != g_ascii_tolower (d[i])) {
+			return s[i] - d[i];
 		}
 
 		leftover--;
-		s++;
-		d++;
+		i++;
 	}
 
 	return ret;
@@ -219,17 +226,17 @@ rspamd_icase_hash (const gchar *in, gsize len, guint64 seed)
 
 	switch (leftover) {
 	case 7:
-		u.c.c7 = lc_map[(guchar)s[i++]]; /* fallthrough */
+		u.c.c7 = lc_map[(guchar)s[i++]]; /* FALLTHRU */
 	case 6:
-		u.c.c6 = lc_map[(guchar)s[i++]]; /* fallthrough */
+		u.c.c6 = lc_map[(guchar)s[i++]]; /* FALLTHRU */
 	case 5:
-		u.c.c5 = lc_map[(guchar)s[i++]]; /* fallthrough */
+		u.c.c5 = lc_map[(guchar)s[i++]]; /* FALLTHRU */
 	case 4:
-		u.c.c4 = lc_map[(guchar)s[i++]]; /* fallthrough */
+		u.c.c4 = lc_map[(guchar)s[i++]]; /* FALLTHRU */
 	case 3:
-		u.c.c3 = lc_map[(guchar)s[i++]]; /* fallthrough */
+		u.c.c3 = lc_map[(guchar)s[i++]]; /* FALLTHRU */
 	case 2:
-		u.c.c2 = lc_map[(guchar)s[i++]]; /* fallthrough */
+		u.c.c2 = lc_map[(guchar)s[i++]]; /* FALLTHRU */
 	case 1:
 		u.c.c1 = lc_map[(guchar)s[i]];
 		break;
@@ -364,29 +371,6 @@ rspamd_strlcpy (gchar *dst, const gchar *src, gsize siz)
 	return (d - dst);
 }
 
-gsize
-rspamd_strlcpy_tolower (gchar *dst, const gchar *src, gsize siz)
-{
-	gchar *d = dst;
-	const gchar *s = src;
-	gsize n = siz;
-
-	/* Copy as many bytes as will fit */
-	if (n != 0) {
-		while (--n != 0) {
-			if ((*d++ = g_ascii_tolower (*s++)) == '\0') {
-				break;
-			}
-		}
-	}
-
-	if (n == 0 && siz != 0) {
-		*d = '\0';
-	}
-
-	return (s - src - 1);    /* count does not include NUL */
-}
-
 /*
  * Try to convert string of length to long
  */
@@ -934,12 +918,15 @@ GString *
 rspamd_header_value_fold (const gchar *name,
 		const gchar *value,
 		guint fold_max,
-		enum rspamd_newlines_type how)
+		enum rspamd_newlines_type how,
+		const gchar *fold_on_chars)
 {
 	GString *res;
 	const guint default_fold_max = 76;
 	guint cur_len;
 	const gchar *p, *c;
+	guint nspaces = 0;
+	const gchar *last;
 	gboolean first_token = TRUE;
 	enum {
 		fold_before = 0,
@@ -969,83 +956,98 @@ rspamd_header_value_fold (const gchar *name,
 
 	while (*p) {
 		switch (state) {
+
 		case read_token:
-			if (*p == ',' || *p == ';') {
-				/* We have something similar to the token's end, so check len */
-				if (cur_len > fold_max * 0.8 && cur_len < fold_max) {
-					/* We want fold */
+			if (fold_on_chars) {
+				if (strchr (fold_on_chars, *p) != NULL) {
 					fold_type = fold_after;
 					state = fold_token;
 					next_state = read_token;
 				}
-				else if (cur_len > fold_max && !first_token) {
-					fold_type = fold_before;
-					state = fold_token;
-					next_state = read_token;
-				}
-				else {
-					g_string_append_len (res, c, p - c + 1);
-					c = p + 1;
-					first_token = FALSE;
-				}
+
 				p ++;
 			}
-			else if (*p == '"') {
-				/* Fold before quoted tokens */
-				g_string_append_len (res, c, p - c);
-				c = p;
-				state = read_quoted;
-			}
-			else if (*p == '\r' || *p == '\n') {
-				if (cur_len > fold_max && !first_token) {
-					fold_type = fold_before;
-					state = fold_token;
-					next_state = read_token;
-				}
-				else {
-					/* Reset line length */
-					cur_len = 0;
-
-					while (g_ascii_isspace (*p)) {
-						p ++;
-					}
-
-					g_string_append_len (res, c, p - c);
-					c = p;
-					first_token = TRUE;
-				}
-			}
-			else if (g_ascii_isspace (*p)) {
-				if (cur_len > fold_max * 0.8 && cur_len < fold_max) {
-					/* We want fold */
-					fold_type = fold_after;
-					state = fold_token;
-					next_state = read_token;
-				}
-				else if (cur_len > fold_max && !first_token) {
-					fold_type = fold_before;
-					state = fold_token;
-					next_state = read_token;
-				}
-				else {
-					g_string_append_len (res, c, p - c);
-					c = p;
-					first_token = FALSE;
-					p ++;
-					cur_len ++;
-				}
-			}
 			else {
-				p ++;
-				cur_len ++;
+				if (*p == ',' || *p == ';') {
+					/* We have something similar to the token's end, so check len */
+					if (cur_len > fold_max * 0.8 && cur_len < fold_max) {
+						/* We want fold */
+						fold_type = fold_after;
+						state = fold_token;
+						next_state = read_token;
+					} else if (cur_len > fold_max && !first_token) {
+						fold_type = fold_before;
+						state = fold_token;
+						next_state = read_token;
+					} else {
+						g_string_append_len (res, c, p - c + 1);
+						c = p + 1;
+						first_token = FALSE;
+					}
+					p++;
+				} else if (*p == '"') {
+					/* Fold before quoted tokens */
+					g_string_append_len (res, c, p - c);
+					c = p;
+					state = read_quoted;
+				} else if (*p == '\r' || *p == '\n') {
+					if (cur_len > fold_max && !first_token) {
+						fold_type = fold_before;
+						state = fold_token;
+						next_state = read_token;
+					} else {
+						/* Reset line length */
+						cur_len = 0;
+
+						while (g_ascii_isspace (*p)) {
+							p++;
+						}
+
+						g_string_append_len (res, c, p - c);
+						c = p;
+						first_token = TRUE;
+					}
+				} else if (g_ascii_isspace (*p)) {
+					if (cur_len > fold_max * 0.8 && cur_len < fold_max) {
+						/* We want fold */
+						fold_type = fold_after;
+						state = fold_token;
+						next_state = read_token;
+					} else if (cur_len > fold_max && !first_token) {
+						fold_type = fold_before;
+						state = fold_token;
+						next_state = read_token;
+					} else {
+						g_string_append_len (res, c, p - c);
+						c = p;
+						first_token = FALSE;
+						p++;
+						cur_len++;
+					}
+				} else {
+					p++;
+					cur_len++;
+				}
 			}
 			break;
 		case fold_token:
 			/* Here, we have token start at 'c' and token end at 'p' */
 			if (fold_type == fold_after) {
-
+				nspaces = 0;
 				if (p > c) {
 					g_string_append_len (res, c, p - c);
+
+					/*
+					 * Check any spaces that are appended to the result
+					 * before folding
+					 */
+					last = &res->str[res->len - 1];
+
+					while (g_ascii_isspace (*last)) {
+						last --;
+						nspaces ++;
+						res->len --;
+					}
 				}
 
 				switch (how) {
@@ -1066,30 +1068,59 @@ rspamd_header_value_fold (const gchar *name,
 					p ++;
 				}
 
+				/* Move leftover spaces */
+				while (nspaces) {
+					g_string_append_c (res, ' ');
+					nspaces --;
+				}
+
 				cur_len = 0;
 			}
 			else {
+				const gchar *last;
+
 				/* Skip space if needed */
 				if (g_ascii_isspace (*c) && p > c) {
 					c ++;
 				}
 
-				switch (how) {
-				case RSPAMD_TASK_NEWLINES_LF:
-					g_string_append_len (res, "\n\t", 2);
-					break;
-				case RSPAMD_TASK_NEWLINES_CR:
-					g_string_append_len (res, "\r\t", 2);
-					break;
-				case RSPAMD_TASK_NEWLINES_CRLF:
-				default:
-					g_string_append_len (res, "\r\n\t", 3);
-					break;
+				/* Avoid double folding */
+				last = &res->str[res->len - 1];
+				last --;
+
+				if (*last != '\r' && *last != '\n') {
+					last ++;
+					while (g_ascii_isspace (*last)) {
+						last --;
+						nspaces ++;
+						res->len --;
+					}
+
+					switch (how) {
+					case RSPAMD_TASK_NEWLINES_LF:
+						g_string_append_len (res, "\n\t", 2);
+						break;
+					case RSPAMD_TASK_NEWLINES_CR:
+						g_string_append_len (res, "\r\t", 2);
+						break;
+					case RSPAMD_TASK_NEWLINES_CRLF:
+					default:
+						g_string_append_len (res, "\r\n\t", 3);
+						break;
+					}
+				}
+
+				/* Move leftover spaces */
+				cur_len = nspaces;
+
+				while (nspaces) {
+					g_string_append_c (res, ' ');
+					nspaces --;
 				}
 
 				if (p > c) {
 					g_string_append_len (res, c, p - c);
-					cur_len = p - c;
+					cur_len += p - c;
 				}
 				else {
 					cur_len = 0;
@@ -1124,7 +1155,7 @@ rspamd_header_value_fold (const gchar *name,
 	/* Last token */
 	switch (state) {
 	case read_token:
-		if (cur_len > fold_max && !first_token) {
+		if (!fold_on_chars && cur_len > fold_max && !first_token) {
 			if (g_ascii_isspace (*c)) {
 				c ++;
 			}
@@ -1159,10 +1190,8 @@ rspamd_header_value_fold (const gchar *name,
 	return res;
 }
 
-static inline bool rspamd_substring_cmp_func (guchar a, guchar b) RSPAMD_ALWAYS_INLINE;
 static inline bool rspamd_substring_cmp_func (guchar a, guchar b) { return a == b; }
 
-static inline bool rspamd_substring_casecmp_func (guchar a, guchar b) RSPAMD_ALWAYS_INLINE;
 static inline bool rspamd_substring_casecmp_func (guchar a, guchar b) { return lc_map[a] == lc_map[b]; }
 
 typedef bool (*rspamd_cmpchar_func_t) (guchar a, guchar b);
@@ -1620,9 +1649,9 @@ decode:
 				c = *p++;
 				ret *= 16;
 
-				if      (c >= '0' && c <= '9') ret += c - '0';
-				else if (c >= 'A' && c <= 'F') ret += c - 'A' + 10;
-				else if (c >= 'a' && c <= 'f') ret += c - 'a' + 10;
+				if      (c >= '0' && c <= '9') { ret += c - '0'; }
+				else if (c >= 'A' && c <= 'F') { ret += c - 'A' + 10; }
+				else if (c >= 'a' && c <= 'f') { ret += c - 'a' + 10; }
 
 				if (end - o > 0) {
 					*o++ = (gchar)ret;
@@ -1724,9 +1753,9 @@ decode:
 				c = *p++;
 				ret *= 16;
 
-				if      (c >= '0' && c <= '9') ret += c - '0';
-				else if (c >= 'A' && c <= 'F') ret += c - 'A' + 10;
-				else if (c >= 'a' && c <= 'f') ret += c - 'a' + 10;
+				if      (c >= '0' && c <= '9') { ret += c - '0'; }
+				else if (c >= 'A' && c <= 'F') { ret += c - 'A' + 10; }
+				else if (c >= 'a' && c <= 'f') { ret += c - 'a' + 10; }
 
 				if (end - o > 0) {
 					*o++ = (gchar)ret;
@@ -1862,15 +1891,18 @@ rspamd_gstring_append_double (double val, void *ud)
 	GString *buf = ud;
 	const double delta = 0.0000001;
 
-	if (val == (double) (int) val) {
-		rspamd_printf_gstring (buf, "%.1f", val);
-	}
-	else if (fabs (val - (double) (int) val) < delta) {
-		/* Write at maximum precision */
-		rspamd_printf_gstring (buf, "%.*g", DBL_DIG, val);
+	if (isfinite (val)) {
+		if (val == (double) (int) val) {
+			rspamd_printf_gstring (buf, "%.1f", val);
+		} else if (fabs (val - (double) (int) val) < delta) {
+			/* Write at maximum precision */
+			rspamd_printf_gstring (buf, "%.*g", DBL_DIG, val);
+		} else {
+			rspamd_printf_gstring (buf, "%f", val);
+		}
 	}
 	else {
-		rspamd_printf_gstring (buf, "%f", val);
+		rspamd_printf_gstring (buf, "null");
 	}
 
 	return 0;
@@ -1929,17 +1961,18 @@ static int
 rspamd_fstring_emit_append_double (double val, void *ud)
 {
 	rspamd_fstring_t **buf = ud;
-	const double delta = 0.0000001;
+#define MAX_PRECISION 6
 
-	if (val == (double)((gint) val)) {
-		rspamd_printf_fstring (buf, "%.1f", val);
-	}
-	else if (fabs (val - (double) (int) val) < delta) {
-		/* Write at maximum precision */
-		rspamd_printf_fstring (buf, "%.*g", DBL_DIG, val);
+	if (isfinite (val)) {
+		if (val == (double) ((gint) val)) {
+			rspamd_printf_fstring (buf, "%.1f", val);
+		} else {
+			rspamd_printf_fstring (buf, "%." G_STRINGIFY (MAX_PRECISION) "f",
+					val);
+		}
 	}
 	else {
-		rspamd_printf_fstring (buf, "%f", val);
+		rspamd_printf_fstring (buf, "null");
 	}
 
 	return 0;
@@ -1976,3 +2009,208 @@ rspamd_memrchr (const void *m, gint c, gsize len)
 
 	return NULL;
 }
+
+gboolean
+rspamd_normalise_unicode_inplace (rspamd_mempool_t *pool, gchar *start,
+		guint *len)
+{
+#if U_ICU_VERSION_MAJOR_NUM >= 44
+	UErrorCode uc_err = U_ZERO_ERROR;
+	static UConverter *utf8_conv = NULL;
+	static const UNormalizer2 *norm = NULL;
+	gint32 nsym, end;
+	UChar *src = NULL, *dest = NULL;
+	gboolean ret = FALSE;
+
+	if (utf8_conv == NULL) {
+		utf8_conv = ucnv_open ("UTF-8", &uc_err);
+		g_assert (U_SUCCESS (uc_err));
+		norm = unorm2_getInstance (NULL, "nfkc", UNORM2_COMPOSE, &uc_err);
+		g_assert (U_SUCCESS (uc_err));
+	}
+
+	/* We first need to convert data to UChars :( */
+	src = g_malloc ((*len + 1) * sizeof (*src));
+	nsym = ucnv_toUChars (utf8_conv, src, *len + 1,
+			start, *len, &uc_err);
+
+	if (!U_SUCCESS (uc_err)) {
+		msg_warn_pool_check ("cannot normalise URL, cannot convert to unicode: %s",
+				u_errorName (uc_err));
+		goto out;
+	}
+
+	/* We can now check if we need to decompose */
+	end = unorm2_spanQuickCheckYes (norm, src, nsym, &uc_err);
+
+	if (!U_SUCCESS (uc_err)) {
+		msg_warn_pool_check ("cannot normalise URL, cannot check normalisation: %s",
+				u_errorName (uc_err));
+		goto out;
+	}
+
+	if (end == nsym) {
+		/* No normalisation needed */
+		goto out;
+	}
+
+	/* We copy sub(src, 0, end) to dest and normalise the rest */
+	ret = TRUE;
+	dest = g_malloc (nsym * sizeof (*dest));
+	memcpy (dest, src, end * sizeof (*dest));
+	nsym = unorm2_normalizeSecondAndAppend (norm, dest, end, nsym,
+			src + end, nsym - end, &uc_err);
+
+	if (!U_SUCCESS (uc_err)) {
+		if (uc_err != U_BUFFER_OVERFLOW_ERROR) {
+			msg_warn_pool_check ("cannot normalise URL: %s",
+					u_errorName (uc_err));
+		}
+
+		goto out;
+	}
+
+	/* We now convert it back to utf */
+	nsym = ucnv_fromUChars (utf8_conv, start, *len, dest, nsym, &uc_err);
+
+	if (!U_SUCCESS (uc_err)) {
+		msg_warn_pool_check ("cannot normalise URL, cannot convert to UTF8: %s",
+				u_errorName (uc_err));
+		goto out;
+	}
+
+	*len = nsym;
+
+out:
+
+	if (src) {
+		g_free (src);
+	}
+
+	if (dest) {
+		g_free (dest);
+	}
+
+	return ret;
+#else
+	/* Kill that with fire please */
+	return FALSE;
+#endif
+}
+
+gchar *
+rspamd_str_regexp_escape (const gchar *pattern, gsize slen,
+		gsize *dst_len, gboolean allow_glob)
+{
+	const gchar *p, *end = pattern + slen;
+	gchar *res, *d, t;
+	gsize len;
+	static const gchar hexdigests[16] = "0123456789abcdef";
+
+	len = slen;
+	p = pattern;
+
+	/* [-[\]{}()*+?.,\\^$|#\s] need to be escaped */
+	while (p < end) {
+		t = *p ++;
+
+		switch (t) {
+		case '[':
+		case ']':
+		case '-':
+		case '\\':
+		case '{':
+		case '}':
+		case '(':
+		case ')':
+		case '*':
+		case '+':
+		case '?':
+		case '.':
+		case ',':
+		case '^':
+		case '$':
+		case '|':
+		case '#':
+			len ++;
+			break;
+		default:
+			if (g_ascii_isspace (t)) {
+				len ++;
+			}
+			else if (!g_ascii_isprint (t)) {
+				/* \\xHH -> 4 symbols */
+				len += 3;
+			}
+			break;
+		}
+	}
+
+	if (slen == len) {
+		if (dst_len) {
+			*dst_len = slen;
+		}
+
+		return g_strdup (pattern);
+	}
+
+	res = g_malloc (len + 1);
+	p = pattern;
+	d = res;
+
+	while (p < end) {
+		t = *p ++;
+
+		switch (t) {
+		case '[':
+		case ']':
+		case '-':
+		case '\\':
+		case '{':
+		case '}':
+		case '(':
+		case ')':
+		case '.':
+		case ',':
+		case '^':
+		case '$':
+		case '|':
+		case '#':
+			*d++ = '\\';
+			break;
+		case '*':
+		case '?':
+		case '+':
+			if (allow_glob) {
+				/* Treat * as .* and ? as .? */
+				*d++ = '.';
+			}
+			else {
+				*d++ = '\\';
+			}
+			break;
+		default:
+			if (g_ascii_isspace (t)) {
+				*d++ = '\\';
+			}
+			else if (!g_ascii_isgraph (t)) {
+				*d++ = '\\';
+				*d++ = 'x';
+				*d++ = hexdigests[((t >> 4) & 0xF)];
+				*d++ = hexdigests[((t) & 0xF)];
+				continue; /* To avoid *d++ = t; */
+			}
+			break;
+		}
+
+		*d++ = t;
+	}
+
+	*d = '\0';
+
+	if (dst_len) {
+		*dst_len = d - res;
+	}
+
+	return res;
+}
