@@ -19,6 +19,7 @@
 #include "unix-std.h"
 #include "contrib/zstd/zstd.h"
 #include "libmime/email_addr.h"
+#include "linenoise.h"
 #include <math.h>
 #include <glob.h>
 
@@ -49,7 +50,7 @@ LUA_FUNCTION_DEF (util, load_rspamd_config);
  */
 LUA_FUNCTION_DEF (util, config_from_ucl);
 /***
- * @function util.encode_base64(input[, str_len])
+ * @function util.encode_base64(input[, str_len, [newlines_type]])
  * Encodes data in base64 breaking lines if needed
  * @param {text or string} input input data
  * @param {number} str_len optional size of lines or 0 if split is not needed
@@ -134,11 +135,13 @@ LUA_FUNCTION_DEF (util, levenshtein_distance);
 LUA_FUNCTION_DEF (util, parse_addr);
 
 /***
- * @function util.fold_header(name, value)
+ * @function util.fold_header(name, value, [how, [stop_chars]])
  * Fold rfc822 header according to the folding rules
  *
  * @param {string} name name of the header
  * @param {string} value value of the header
+ * @param {string} how "cr" for \r, "lf" for \n and "crlf" for \r\n (default)
+ * @param {string} stop_chars also fold header when the
  * @return {string} Folded value of the header
  */
 LUA_FUNCTION_DEF (util, fold_header);
@@ -180,7 +183,7 @@ LUA_FUNCTION_DEF (util, get_tld);
 LUA_FUNCTION_DEF (util, glob);
 
 /***
- * @function util.parse_mail_address(str)
+ * @function util.parse_mail_address(str, pool)
  * Parses email address and returns a table of tables in the following format:
  *
  * - `name` - name of internet address in UTF8, e.g. for `Vsevolod Stakhov <blah@foo.com>` it returns `Vsevolod Stakhov`
@@ -189,6 +192,7 @@ LUA_FUNCTION_DEF (util, glob);
  * - `domain` - domain part (if present), e.g. `foo.com`
  *
  * @param {string} str input string
+ * @param {rspamd_mempool} pool memory pool to use
  * @return {table/tables} parsed list of mail addresses
  */
 LUA_FUNCTION_DEF (util, parse_mail_address);
@@ -374,6 +378,49 @@ LUA_FUNCTION_DEF (util, is_utf_spoofed);
  */
 LUA_FUNCTION_DEF (util, is_valid_utf8);
 
+/***
+ * @function util.readline([prompt])
+ * Returns string read from stdin with history and editing support
+ * @return {string} string read from the input (with line endings stripped)
+ */
+LUA_FUNCTION_DEF (util, readline);
+
+/***
+ * @function util.readpassphrase([prompt])
+ * Returns string read from stdin disabling echo
+ * @return {string} string read from the input (with line endings stripped)
+ */
+LUA_FUNCTION_DEF (util, readpassphrase);
+
+/***
+ * @function util.file_exists(file)
+ * Checks if a specified file exists and is available for reading
+ * @return {boolean} true if file exists
+ */
+LUA_FUNCTION_DEF (util, file_exists);
+
+/***
+ * @function util.mkdir(dir[, recursive])
+ * Creates a specified directory
+ * @return {boolean[,error]} true if directory has been created
+ */
+LUA_FUNCTION_DEF (util, mkdir);
+
+/***
+ * @function util.umask(mask)
+ * Sets new umask. Accepts either numeric octal string, e.g. '022' or a plain
+ * number, e.g. 0x12 (since Lua does not support octal integrals)
+ * @return {number} old umask
+ */
+LUA_FUNCTION_DEF (util, umask);
+
+/***
+ * @function util.isatty()
+ * Returns if stdout is a tty
+ * @return {boolean} true in case of output being tty
+ */
+LUA_FUNCTION_DEF (util, isatty);
+
 /***
  * @function util.pack(fmt, ...)
  *
@@ -473,6 +520,7 @@ LUA_FUNCTION_DEF (util, caseless_hash_fast);
  */
 LUA_FUNCTION_DEF (util, get_hostname);
 
+
 static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, create_event_base),
 	LUA_INTERFACE_DEF (util, load_rspamd_config),
@@ -515,6 +563,12 @@ static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, caseless_hash_fast),
 	LUA_INTERFACE_DEF (util, is_utf_spoofed),
 	LUA_INTERFACE_DEF (util, is_valid_utf8),
+	LUA_INTERFACE_DEF (util, readline),
+	LUA_INTERFACE_DEF (util, readpassphrase),
+	LUA_INTERFACE_DEF (util, file_exists),
+	LUA_INTERFACE_DEF (util, mkdir),
+	LUA_INTERFACE_DEF (util, umask),
+	LUA_INTERFACE_DEF (util, isatty),
 	LUA_INTERFACE_DEF (util, get_hostname),
 	LUA_INTERFACE_DEF (util, pack),
 	LUA_INTERFACE_DEF (util, unpack),
@@ -648,8 +702,8 @@ lua_util_process_message (lua_State *L)
 
 	if (cfg != NULL && message != NULL) {
 		base = event_init ();
-		rspamd_init_filters (cfg, FALSE, NULL);
-		task = rspamd_task_new (NULL, cfg, NULL);
+		rspamd_init_filters (cfg, FALSE);
+		task = rspamd_task_new (NULL, cfg, NULL, NULL);
 		task->ev_base = base;
 		task->msg.begin = rspamd_mempool_alloc (task->task_pool, mlen);
 		rspamd_strlcpy ((gpointer)task->msg.begin, message, mlen);
@@ -702,6 +756,7 @@ lua_util_encode_base64 (lua_State *L)
 	gchar *out;
 	gsize inlen, outlen;
 	guint str_lim = 0;
+	gboolean fold = FALSE;
 
 	if (lua_type (L, 1) == LUA_TSTRING) {
 		s = luaL_checklstring (L, 1, &inlen);
@@ -717,13 +772,37 @@ lua_util_encode_base64 (lua_State *L)
 
 	if (lua_gettop (L) > 1) {
 		str_lim = luaL_checknumber (L, 2);
+
+		fold = !!(str_lim > 0);
 	}
 
 	if (s == NULL) {
 		lua_pushnil (L);
 	}
 	else {
-		out = rspamd_encode_base64 (s, inlen, str_lim, &outlen);
+
+		if (fold) {
+			out = rspamd_encode_base64 (s, inlen, str_lim, &outlen);
+		}
+		else {
+			enum rspamd_newlines_type how = RSPAMD_TASK_NEWLINES_CRLF;
+
+			if (lua_type (L, 3) == LUA_TSTRING) {
+				const gchar *how_str = lua_tostring (L, 3);
+
+				if (g_ascii_strcasecmp (how_str, "cr") == 0) {
+					how = RSPAMD_TASK_NEWLINES_CR;
+				}
+				else if (g_ascii_strcasecmp (how_str, "lf") == 0) {
+					how = RSPAMD_TASK_NEWLINES_LF;
+				}
+				else if (g_ascii_strcasecmp (how_str, "crlf") != 0) {
+					return luaL_error (L, "invalid newline style: %s", how_str);
+				}
+			}
+
+			out = rspamd_encode_base64_fold (s, inlen, str_lim, &outlen, how);
+		}
 
 		if (out != NULL) {
 			t = lua_newuserdata (L, sizeof (*t));
@@ -951,9 +1030,10 @@ lua_util_tokenize_text (lua_State *L)
 				lua_pop (L, 1);
 
 				if (ex_len > 0) {
-					ex = g_slice_alloc (sizeof (*ex));
+					ex = g_malloc0 (sizeof (*ex));
 					ex->pos = pos;
 					ex->len = ex_len;
+					ex->type = RSPAMD_EXCEPTION_URL;
 					exceptions = g_list_prepend (exceptions, ex);
 				}
 			}
@@ -990,7 +1070,7 @@ lua_util_tokenize_text (lua_State *L)
 	cur = exceptions;
 	while (cur) {
 		ex = cur->data;
-		g_slice_free1 (sizeof (*ex), ex);
+		g_free (ex);
 		cur = g_list_next (cur);
 	}
 
@@ -1127,7 +1207,7 @@ lua_util_parse_addr (lua_State *L)
 static gint
 lua_util_fold_header (lua_State *L)
 {
-	const gchar *name, *value, *how;
+	const gchar *name, *value, *how, *stop_chars = NULL;
 	GString *folded;
 
 	name = luaL_checkstring (L, 1);
@@ -1136,24 +1216,29 @@ lua_util_fold_header (lua_State *L)
 	if (name && value) {
 
 		if (lua_isstring (L, 3)) {
+
 			how = lua_tostring (L, 3);
 
+			if (lua_isstring (L, 4)) {
+				stop_chars = lua_tostring (L, 4);
+			}
+
 			if (strcmp (how, "cr") == 0) {
 				folded = rspamd_header_value_fold (name, value, 0,
-						RSPAMD_TASK_NEWLINES_CR);
+						RSPAMD_TASK_NEWLINES_CR, stop_chars);
 			}
 			else if (strcmp (how, "lf") == 0) {
 				folded = rspamd_header_value_fold (name, value, 0,
-						RSPAMD_TASK_NEWLINES_LF);
+						RSPAMD_TASK_NEWLINES_LF, stop_chars);
 			}
 			else {
 				folded = rspamd_header_value_fold (name, value, 0,
-						RSPAMD_TASK_NEWLINES_CRLF);
+						RSPAMD_TASK_NEWLINES_CRLF, stop_chars);
 			}
 		}
 		else {
 			folded = rspamd_header_value_fold (name, value, 0,
-					RSPAMD_TASK_NEWLINES_CRLF);
+					RSPAMD_TASK_NEWLINES_CRLF, stop_chars);
 		}
 
 		if (folded) {
@@ -1171,33 +1256,28 @@ lua_util_fold_header (lua_State *L)
 static gint
 lua_util_is_uppercase (lua_State *L)
 {
-	const gchar *str, *p;
-	gsize sz, remain;
-	gunichar uc;
+	const gchar *str;
+	gsize sz;
+	gint32 i = 0;
+	UChar32 uc;
 	guint nlc = 0, nuc = 0;
 
 	str = luaL_checklstring (L, 1, &sz);
-	remain = sz;
 
-	if (str && remain > 0) {
-		while (remain > 0) {
-			uc = g_utf8_get_char_validated (str, remain);
-			p = g_utf8_next_char (str);
+	if (str && sz > 0) {
+		while (i >= 0 && i < sz) {
+			U8_NEXT (str, i, sz, uc);
 
-			if (p - str > (gint) remain) {
+			if (uc < 0) {
 				break;
 			}
 
-			remain -= p - str;
-
-			if (g_unichar_isupper (uc)) {
+			if (u_isupper (uc)) {
 				nuc++;
 			}
-			else if (g_unichar_islower (uc)) {
+			else if (u_islower (uc)) {
 				nlc++;
 			}
-
-			str = p;
 		}
 	}
 
@@ -1434,8 +1514,13 @@ static gint
 lua_util_get_ticks (lua_State *L)
 {
 	gdouble ticks;
+	gboolean rdtsc = FALSE;
 
-	ticks = rspamd_get_ticks ();
+	if (lua_isboolean (L, 1)) {
+		rdtsc = lua_toboolean (L, 1);
+	}
+
+	ticks = rspamd_get_ticks (rdtsc);
 	lua_pushnumber (L, ticks);
 
 	return 1;
@@ -1563,6 +1648,15 @@ lua_util_lock_file (lua_State *L)
 	gint fd = -1;
 	gboolean own = FALSE;
 
+#if !HAVE_FLOCK
+	struct flock fl = {
+		.l_type = F_WRLCK,
+		.l_whence = SEEK_SET,
+		.l_start = 0,
+		.l_len = 0
+	};
+#endif
+
 	fpath = luaL_checkstring (L, 1);
 
 	if (fpath) {
@@ -1581,7 +1675,11 @@ lua_util_lock_file (lua_State *L)
 			return 2;
 		}
 
+#if HAVE_FLOCK
 		if (flock (fd, LOCK_EX) == -1) {
+#else
+		if (fcntl (fd, F_SETLKW, &fl) == -1) {
+#endif
 			lua_pushnil (L);
 			lua_pushstring (L, strerror (errno));
 
@@ -1607,6 +1705,15 @@ lua_util_unlock_file (lua_State *L)
 	gint fd = -1, ret, serrno;
 	gboolean do_close = TRUE;
 
+#if !HAVE_FLOCK
+	struct flock fl = {
+		.l_type = F_UNLCK,
+		.l_whence = SEEK_SET,
+		.l_start = 0,
+		.l_len = 0
+	};
+#endif
+
 	if (lua_isnumber (L, 1)) {
 		fd = lua_tonumber (L, 1);
 
@@ -1614,7 +1721,11 @@ lua_util_unlock_file (lua_State *L)
 			do_close = lua_toboolean (L, 2);
 		}
 
+#if HAVE_FLOCK
 		ret = flock (fd, LOCK_UN);
+#else
+		ret = fcntl (fd, F_SETLKW, &fl);
+#endif
 
 		if (do_close) {
 			serrno = errno;
@@ -1716,11 +1827,11 @@ lua_util_random_hex (lua_State *L)
 static gint
 lua_util_zstd_compress (lua_State *L)
 {
-	struct rspamd_lua_text *t = NULL, *res;
+	struct rspamd_lua_text *t = NULL, *res, tmp;
 	gsize sz, r;
 
 	if (lua_type (L, 1) == LUA_TSTRING) {
-		t = g_alloca (sizeof (*t));
+		t = &tmp;
 		t->start = lua_tolstring (L, 1, &sz);
 		t->len = sz;
 	}
@@ -2050,6 +2161,173 @@ lua_util_is_valid_utf8 (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_util_readline (lua_State *L)
+{
+	const gchar *prompt = NULL;
+	gchar *input;
+
+	if (lua_type (L, 1) == LUA_TSTRING) {
+		prompt = lua_tostring (L, 1);
+	}
+
+	input = linenoise (prompt);
+
+	if (input) {
+		lua_pushstring (L, input);
+		linenoiseHistoryAdd (input);
+		linenoiseFree (input);
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	return 1;
+}
+
+static gint
+lua_util_readpassphrase (lua_State *L)
+{
+	gchar test_password[8192];
+	gsize r;
+
+	r = rspamd_read_passphrase (test_password, sizeof (test_password), 0, NULL);
+
+	if (r > 0) {
+		lua_pushlstring (L, test_password, r);
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	/* In fact, we still pass it to Lua which is not very safe */
+	rspamd_explicit_memzero (test_password, sizeof (test_password));
+
+	return 1;
+}
+
+static gint
+lua_util_file_exists (lua_State *L)
+{
+	const gchar *fname = luaL_checkstring (L, 1);
+
+	if (fname) {
+		lua_pushboolean (L, access (fname, R_OK) != -1);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+static gint
+lua_util_mkdir (lua_State *L)
+{
+	const gchar *dname = luaL_checkstring (L, 1);
+	gboolean recursive = FALSE;
+	gint r = -1;
+
+	if (dname) {
+		if (lua_isboolean (L, 2)) {
+			recursive = lua_toboolean (L, 2);
+		}
+
+		if (recursive) {
+			char path[PATH_MAX];
+			gsize len, i;
+
+			len = rspamd_strlcpy (path, dname, sizeof (path));
+
+			/* Strip last / */
+			if (path[len - 1] == '/') {
+				path[len - 1] = '\0';
+				len --;
+			}
+
+			for (i = 1; i < len; i ++) {
+				if (path[i] == '/') {
+					path[i] = '\0';
+
+					errno = 0;
+					r = mkdir (path, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+
+					if (r == -1 && errno != EEXIST) {
+						break;
+					}
+
+					path[i] = '/';
+				}
+			}
+
+			/* Final path component */
+			r = mkdir (path, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+		}
+		else {
+			r = mkdir (dname, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+		}
+
+		if (r == -1 && errno != EEXIST) {
+			lua_pushboolean (L, false);
+			lua_pushstring (L, strerror (errno));
+
+			return 2;
+		}
+
+		lua_pushboolean (L, true);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+
+static gint
+lua_util_umask (lua_State *L)
+{
+	mode_t mask = 0, old;
+
+	if (lua_type (L, 1) == LUA_TSTRING) {
+		const gchar *str = lua_tostring (L, 1);
+
+		if (str[0] == '0') {
+			/* e.g. '022' */
+			mask = strtol (str, NULL, 8);
+		}
+		else {
+			/* XXX: implement modestring parsing at some point */
+			return luaL_error (L, "invalid arguments");
+		}
+	}
+	else if (lua_type (L, 1) == LUA_TNUMBER) {
+		mask = lua_tonumber (L, 1);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	old = umask (mask);
+
+	lua_pushnumber (L, old);
+
+	return 1;
+}
+
+static gint
+lua_util_isatty (lua_State *L)
+{
+	if (isatty (STDOUT_FILENO)) {
+		lua_pushboolean (L, true);
+	}
+	else {
+		lua_pushboolean (L, false);
+	}
+
+	return 1;
+}
+
 /* Backport from Lua 5.3 */
 
 /******************************************************************************
@@ -2222,14 +2500,19 @@ COMPAT53_API void
 luaL_buffinit (lua_State *L, luaL_Buffer_53 *B)
 {
 	/* make it crash if used via pointer to a 5.1-style luaL_Buffer */
+#if LUA_VERSION_NUM < 502
 	B->b.p = NULL;
 	B->b.L = NULL;
 	B->b.lvl = 0;
 	/* reuse the buffer from the 5.1-style luaL_Buffer though! */
 	B->ptr = B->b.buffer;
-	B->capacity = LUAL_BUFFERSIZE;
 	B->nelems = 0;
-	B->L2 = L;
+#elif LUA_VERSION_NUM == 502
+        B->ptr = B->b.b;
+	B->nelems = B->b.n;
+#endif
+	B->capacity = LUAL_BUFFERSIZE;
+        B->L2 = L;
 }
 
 
@@ -2245,7 +2528,11 @@ luaL_prepbuffsize (luaL_Buffer_53 *B, size_t s)
 			luaL_error (B->L2, "buffer too large");
 		newptr = (char *) lua_newuserdata (B->L2, newcap);
 		memcpy(newptr, B->ptr, B->nelems);
+#if LUA_VERSION_NUM < 502
 		if (B->ptr != B->b.buffer)
+#elif LUA_VERSION_NUM == 502
+		if (B->ptr != B->b.b)
+#endif
 			lua_replace (B->L2, -2); /* remove old buffer */
 		B->ptr = newptr;
 		B->capacity = newcap;
@@ -2269,10 +2556,18 @@ luaL_addvalue (luaL_Buffer_53 *B)
 	const char *s = lua_tolstring (B->L2, -1, &len);
 	if (!s)
 		luaL_error (B->L2, "cannot convert value to string");
+#if LUA_VERSION_NUM < 502
 	if (B->ptr != B->b.buffer)
+#elif LUA_VERSION_NUM == 502
+	if (B->ptr != B->b.b)
+#endif
 		lua_insert (B->L2, -2); /* userdata buffer must be at stack top */
 	luaL_addlstring (B, s, len);
+#if LUA_VERSION_NUM < 502
 	lua_remove (B->L2, B->ptr != B->b.buffer ? -2 : -1);
+#elif LUA_VERSION_NUM == 502
+	lua_remove (B->L2, B->ptr != B->b.b ? -2 : -1);
+#endif
 }
 
 
@@ -2280,7 +2575,11 @@ COMPAT53_API void
 luaL_pushresult (luaL_Buffer_53 *B)
 {
 	lua_pushlstring (B->L2, B->ptr, B->nelems);
+#if LUA_VERSION_NUM < 502
 	if (B->ptr != B->b.buffer)
+#elif LUA_VERSION_NUM == 502
+	if (B->ptr != B->b.b)
+#endif
 		lua_replace (B->L2, -2); /* remove userdata buffer */
 }
 
