@@ -17,6 +17,8 @@ limitations under the License.
 local rspamd_logger = require 'rspamd_logger'
 local rspamd_http = require "rspamd_http"
 local rspamd_lua_utils = require "lua_util"
+local upstream_list = require "rspamd_upstream_list"
+local N = "clickhouse"
 
 if confighelp then
   return
@@ -27,6 +29,7 @@ local E = {}
 local rows = {}
 local attachment_rows = {}
 local urls_rows = {}
+local emails_rows = {}
 local specific_rows = {}
 local asn_rows = {}
 local symbols_rows = {}
@@ -45,22 +48,26 @@ local settings = {
   dkim_reject_symbols = {'R_DKIM_REJECT'},
   dmarc_allow_symbols = {'DMARC_POLICY_ALLOW'},
   dmarc_reject_symbols = {'DMARC_POLICY_REJECT', 'DMARC_POLICY_QUARANTINE'},
+  stop_symbols = {},
   table = 'rspamd',
   attachments_table = 'rspamd_attachments',
   urls_table = 'rspamd_urls',
+  emails_table = 'rspamd_emails',
   symbols_table = 'rspamd_symbols',
+  asn_table = 'rspamd_asn',
   ipmask = 19,
   ipmask6 = 48,
   full_urls = false,
   from_tables = nil,
   enable_symbols = false,
   use_https = false,
+  use_gzip = true,
+  allow_local = false,
 }
 
---[[
-Clickhouse schema:
-
-CREATE TABLE rspamd
+local clickhouse_schema = {
+rspamd = [[
+CREATE TABLE IF NOT EXISTS ${table}
 (
     Date Date,
     TS DateTime,
@@ -77,7 +84,7 @@ CREATE TABLE rspamd
     IsDkim Enum8('reject' = 0, 'allow' = 1, 'unknown' = 2) DEFAULT CAST('unknown' AS Enum8('reject' = 0, 'allow' = 1, 'unknown' = 2)),
     IsDmarc Enum8('reject' = 0, 'allow' = 1, 'unknown' = 2) DEFAULT CAST('unknown' AS Enum8('reject' = 0, 'allow' = 1, 'unknown' = 2)),
     NUrls Int32,
-    Action Enum8('reject' = 0, 'rewrite subject' = 1, 'add header' = 2, 'greylist' = 3, 'no action' = 4) DEFAULT CAST('no action' AS Enum8('reject' = 0, 'rewrite subject' = 1, 'add header' = 2, 'greylist' = 3, 'no action' = 4)),
+    Action Enum8('reject' = 0, 'rewrite subject' = 1, 'add header' = 2, 'greylist' = 3, 'no action' = 4, 'soft reject' = 5) DEFAULT CAST('no action' AS Enum8('reject' = 0, 'rewrite subject' = 1, 'add header' = 2, 'greylist' = 3, 'no action' = 4, 'soft reject' = 5)),
     FromUser String,
     MimeUser String,
     RcptUser String,
@@ -85,8 +92,10 @@ CREATE TABLE rspamd
     ListId String,
     Digest FixedString(32)
 ) ENGINE = MergeTree(Date, (TS, From), 8192)
+]],
 
-CREATE TABLE rspamd_attachments (
+  attachments = [[
+CREATE TABLE IF NOT EXISTS ${attachments_table} (
     Date Date,
     Digest FixedString(32),
     `Attachments.FileName` Array(String),
@@ -94,23 +103,37 @@ CREATE TABLE rspamd_attachments (
     `Attachments.Length` Array(UInt32),
     `Attachments.Digest` Array(FixedString(16))
 ) ENGINE = MergeTree(Date, Digest, 8192)
+]],
 
-CREATE TABLE rspamd_urls (
+  urls = [[
+CREATE TABLE IF NOT EXISTS ${urls_table} (
     Date Date,
     Digest FixedString(32),
     `Urls.Tld` Array(String),
     `Urls.Url` Array(String)
 ) ENGINE = MergeTree(Date, Digest, 8192)
+]],
 
-CREATE TABLE rspamd_asn (
+  emails = [[
+CREATE TABLE IF NOT EXISTS ${emails_table} (
+    Date Date,
+    Digest FixedString(32),
+    Emails Array(String)
+) ENGINE = MergeTree(Date, Digest, 8192)
+]],
+
+  asn = [[
+CREATE TABLE IF NOT EXISTS ${asn_table} (
     Date Date,
     Digest FixedString(32),
     ASN String,
     Country FixedString(2),
     IPNet String
 ) ENGINE = MergeTree(Date, Digest, 8192)
+]],
 
-CREATE TABLE rspamd_symbols (
+  symbols = [[
+CREATE TABLE IF NOT EXISTS ${symbols_table} (
     Date Date,
     Digest FixedString(32),
     `Symbols.Names` Array(String),
@@ -118,6 +141,7 @@ CREATE TABLE rspamd_symbols (
     `Symbols.Options` Array(String)
 ) ENGINE = MergeTree(Date, Digest, 8192)
 ]]
+}
 
 local function clickhouse_main_row(tname)
   local fields = {
@@ -176,6 +200,17 @@ local function clickhouse_urls_row(tname)
   return elt
 end
 
+local function clickhouse_emails_row(tname)
+  local emails_fields = {
+    'Date',
+    'Digest',
+    'Emails',
+  }
+  local elt = string.format('INSERT INTO %s (%s) VALUES ',
+      tname, table.concat(emails_fields, ','))
+  return elt
+end
+
 local function clickhouse_symbols_row(tname)
   local symbols_fields = {
     'Date',
@@ -212,6 +247,10 @@ local function clickhouse_first_row()
     table.insert(urls_rows,
       clickhouse_urls_row(settings['urls_table']))
   end
+  if settings['emails_table'] then
+    table.insert(emails_rows,
+        clickhouse_emails_row(settings['emails_table']))
+  end
   if settings['asn_table'] then
     table.insert(asn_rows,
       clickhouse_asn_row(settings['asn_table']))
@@ -238,22 +277,31 @@ local function clickhouse_check_symbol(task, symbols, need_score)
 end
 
 local function clickhouse_send_data(task)
-  local function http_cb(err_message, code, _, _)
-    if code ~= 200 or err_message then
-      rspamd_logger.errx(task, "cannot send data to clickhouse server %s: %s:%s",
-        settings['server'], code, err_message)
-    else
-      rspamd_logger.infox(task, "sent %s rows to clickhouse server %s",
-        settings['limit'], settings['server'])
+  local upstream = settings.upstream:get_upstream_round_robin()
+  local ip_addr = upstream:get_addr():to_string(true)
+
+  local function gen_http_cb(what, how_many)
+    return function (err_message, code, data, _)
+      if code ~= 200 or err_message then
+        if not err_message then err_message = data end
+        rspamd_logger.errx(task, "cannot send %s data to clickhouse server %s: %s",
+            what, ip_addr, err_message)
+        upstream:fail()
+      else
+        rspamd_logger.infox(task, "sent %s rows of %s to clickhouse server %s",
+            how_many - 1, what, ip_addr)
+        upstream:ok()
+      end
     end
   end
 
   local body = table.concat(rows, ' ')
   if not rspamd_http.request({
       task = task,
-      url = connect_prefix .. settings['server'],
+      url = connect_prefix .. ip_addr,
       body = body,
-      callback = http_cb,
+      callback = gen_http_cb('generic data', #rows),
+      gzip = settings.use_gzip,
       mime_type = 'text/plain',
       timeout = settings['timeout'],
     }) then
@@ -265,9 +313,9 @@ local function clickhouse_send_data(task)
     body = table.concat(attachment_rows, ' ')
     if not rspamd_http.request({
       task = task,
-      url = connect_prefix .. settings['server'],
+      url = connect_prefix .. ip_addr,
       body = body,
-      callback = http_cb,
+      callback = gen_http_cb('attachments data', #attachment_rows),
       mime_type = 'text/plain',
       timeout = settings['timeout'],
     }) then
@@ -279,9 +327,9 @@ local function clickhouse_send_data(task)
     body = table.concat(urls_rows, ' ')
     if not rspamd_http.request({
       task = task,
-      url = connect_prefix .. settings['server'],
+      url = connect_prefix .. ip_addr,
       body = body,
-      callback = http_cb,
+      callback = gen_http_cb('urls data', #urls_rows),
       mime_type = 'text/plain',
       timeout = settings['timeout'],
     }) then
@@ -289,13 +337,27 @@ local function clickhouse_send_data(task)
         settings['server'])
     end
   end
+  if #emails_rows > 1 then
+    body = table.concat(emails_rows, ' ')
+    if not rspamd_http.request({
+      task = task,
+      url = connect_prefix .. ip_addr,
+      body = body,
+      callback = gen_http_cb('emails data', #emails_rows),
+      mime_type = 'text/plain',
+      timeout = settings['timeout'],
+    }) then
+      rspamd_logger.errx(task, "cannot send emails to clickhouse server %s: cannot make request",
+          settings['server'])
+    end
+  end
   if #asn_rows > 1 then
     body = table.concat(asn_rows, ' ')
     if not rspamd_http.request({
       task = task,
-      url = connect_prefix .. settings['server'],
+      url = connect_prefix .. ip_addr,
       body = body,
-      callback = http_cb,
+      callback = gen_http_cb('asn data', #asn_rows),
       mime_type = 'text/plain',
       timeout = settings['timeout'],
     }) then
@@ -308,9 +370,9 @@ local function clickhouse_send_data(task)
     body = table.concat(symbols_rows, ' ')
     if not rspamd_http.request({
       task = task,
-      url = connect_prefix .. settings['server'],
+      url = connect_prefix .. ip_addr,
       body = body,
-      callback = http_cb,
+      callback = gen_http_cb('symbols data', #symbols_rows),
       mime_type = 'text/plain',
       timeout = settings['timeout'],
     }) then
@@ -324,9 +386,9 @@ local function clickhouse_send_data(task)
       body = table.concat(specific, ' ')
       if not rspamd_http.request({
         task = task,
-        url = connect_prefix .. settings['server'],
+        url = connect_prefix .. ip_addr,
         body = body,
-        callback = http_cb,
+        callback = gen_http_cb('domain specific data ('..k..')', #specific),
         mime_type = 'text/plain',
         timeout = settings['timeout'],
       }) then
@@ -346,7 +408,15 @@ local function clickhouse_quote(str)
 end
 
 local function clickhouse_collect(task)
-  if rspamd_lua_utils.is_rspamc_or_controller(task) then return end
+  if not settings.allow_local and rspamd_lua_utils.is_rspamc_or_controller(task) then return end
+
+  for _,sym in ipairs(settings.stop_symbols) do
+    if task:has_symbol(sym) then
+      rspamd_logger.debugm(N, task, 'skip collection as symbol %s has fired', sym)
+      return
+    end
+  end
+
   local from_domain = ''
   local from_user = ''
   if task:has_from('smtp') then
@@ -543,7 +613,7 @@ local function clickhouse_collect(task)
   local urls_tlds = {}
   local urls_urls = {}
   if task:has_urls(false) then
-    for _,u in ipairs(task:get_urls()) do
+    for _,u in ipairs(task:get_urls(false)) do
       table.insert(urls_tlds, string.format("'%s'", clickhouse_quote(u:get_tld())))
       if settings['full_urls'] then
         table.insert(urls_urls, string.format("'%s'",
@@ -563,9 +633,26 @@ local function clickhouse_collect(task)
     table.insert(urls_rows, elt)
   end
 
+  -- Emails step
+  local emails = {}
+  if task:has_urls(true) then
+    for _,u in ipairs(task:get_emails()) do
+      table.insert(emails, string.format("'%s'", clickhouse_quote(
+          string.format('%s@%s', u:get_user(), u:get_host())
+      )))
+    end
+  end
+
+  if #emails > 0 then
+    elt = string.format("(today(),'%s',[%s])",
+        task:get_digest(),
+        table.concat(emails, ','))
+    table.insert(emails_rows, elt)
+  end
+
   -- ASN information
   if settings['asn_table'] then
-    local asn, country, ipnet = 'unknown', 'unknown', 'unknown'
+    local asn, country, ipnet = '--', '--', '--'
     local pool = task:get_mempool()
     ret = pool:get_variable("asn")
     if ret then
@@ -573,7 +660,7 @@ local function clickhouse_collect(task)
     end
     ret = pool:get_variable("country")
     if ret then
-      country = ret
+      country = ret:sub(1, 2)
     end
     ret = pool:get_variable("ipnet")
     if ret then
@@ -611,7 +698,7 @@ local function clickhouse_collect(task)
       table.concat(scores_tab, ','),
       table.concat(options_tab, ','))
 
-    table.insert(urls_rows, elt)
+    table.insert(symbols_rows, elt)
   end
 
   nrows = nrows + 1
@@ -622,6 +709,7 @@ local function clickhouse_collect(task)
     rows = {}
     attachment_rows = {}
     urls_rows = {}
+    emails_rows = {}
     specific_rows = {}
     asn_rows = {}
     symbols_rows = {}
@@ -635,8 +723,9 @@ if opts then
       settings[k] = v
     end
 
-    if not settings['server'] then
+    if not settings['server'] and not settings['servers'] then
       rspamd_logger.infox(rspamd_config, 'no servers are specified, disabling module')
+      rspamd_lua_utils.disable_module(N, "config")
     else
       settings['from_map'] = rspamd_map_add('clickhouse', 'from_tables',
         'regexp', 'clickhouse specific domains')
@@ -644,17 +733,68 @@ if opts then
         connect_prefix = 'https://'
       end
 
+      settings.upstream = upstream_list.create(rspamd_config,
+        settings['server'] or settings['servers'], 8123)
+
+      if not settings.upstream then
+        rspamd_logger.errx('cannot parse clickhouse address: %s',
+            settings['server'] or settings['servers'])
+        rspamd_lua_utils.disable_module(N, "config")
+        return
+      end
+
       clickhouse_first_row()
       rspamd_config:register_symbol({
         name = 'CLICKHOUSE_COLLECT',
-        type = 'postfilter',
+        type = 'idempotent',
         callback = clickhouse_collect,
-        priority = 10
+        priority = 10,
+        flags = 'empty',
       })
       rspamd_config:register_finish_script(function(task)
         if nrows > 0 then
           clickhouse_send_data(task)
         end
       end)
+      -- Create tables on load
+      rspamd_config:add_on_load(function(cfg, ev_base, worker)
+        if worker:is_primary_controller() then
+          local upstreams = settings.upstream:all_upstreams()
+
+          for _,up in ipairs(upstreams) do
+            local ip_addr = up:get_addr():to_string(true)
+
+            local function send_req(elt, sql)
+              local function http_cb(err_message, code, data, _)
+                if code ~= 200 or err_message then
+                  if not err_message then err_message = data end
+                  rspamd_logger.errx(rspamd_config, "cannot create table %s in clickhouse server %s: %s",
+                      elt, ip_addr, err_message)
+                  up:fail()
+                else
+                  up:ok()
+                end
+              end
+
+              if not rspamd_http.request({
+                ev_base = ev_base,
+                config = cfg,
+                url = connect_prefix .. ip_addr,
+                body = sql,
+                callback = http_cb,
+                mime_type = 'text/plain',
+                timeout = settings['timeout'],
+              }) then
+                rspamd_logger.errx(rspamd_config, "cannot create table %s in clickhouse server %s: cannot make request",
+                    elt, ip_addr)
+              end
+            end
+
+            for tab,sql in pairs(clickhouse_schema) do
+              send_req(tab, rspamd_lua_utils.template(sql, settings))
+            end
+          end
+        end
+      end)
     end
 end
