@@ -21,6 +21,9 @@ end
 local rspamd_logger = require "rspamd_logger"
 local rspamd_util = require "rspamd_util"
 local fun = require "fun"
+local lua_util = require "lua_util"
+
+local N = "whitelist"
 
 local options = {
   dmarc_allow_symbol = 'DMARC_POLICY_ALLOW',
@@ -252,7 +255,7 @@ local configure_whitelist_module = function()
           flags = 'empty'
         end
 
-        local id = rspamd_config:register_symbol({
+        rspamd_config:register_symbol({
           name = symbol,
           flags = flags,
           callback = gen_whitelist_cb(symbol, rule)
@@ -261,21 +264,21 @@ local configure_whitelist_module = function()
         local spf_dep = false
         local dkim_dep = false
         if rule['valid_spf'] then
-          rspamd_config:register_dependency(id, options['spf_allow_symbol'])
+          rspamd_config:register_dependency(symbol, options['spf_allow_symbol'])
           spf_dep = true
         end
         if rule['valid_dkim'] then
-          rspamd_config:register_dependency(id, options['dkim_allow_symbol'])
+          rspamd_config:register_dependency(symbol, options['dkim_allow_symbol'])
           dkim_dep = true
         end
         if rule['valid_dmarc'] then
           if not spf_dep then
-            rspamd_config:register_dependency(id, options['spf_allow_symbol'])
+            rspamd_config:register_dependency(symbol, options['spf_allow_symbol'])
           end
           if not dkim_dep then
-            rspamd_config:register_dependency(id, options['dkim_allow_symbol'])
+            rspamd_config:register_dependency(symbol, options['dkim_allow_symbol'])
           end
-          rspamd_config:register_dependency(id, 'DMARC_CALLBACK')
+          rspamd_config:register_dependency(symbol, 'DMARC_CALLBACK')
         end
 
         if rule['score'] then
@@ -287,6 +290,8 @@ local configure_whitelist_module = function()
         end
       end
     end, options['rules'])
+  else
+    lua_util.disable_module(N, "config")
   end
 end
 
