@@ -54,7 +54,8 @@ struct rspamadm_command lua_command = {
 		.name = "lua",
 		.flags = 0,
 		.help = rspamadm_lua_help,
-		.run = rspamadm_lua
+		.run = rspamadm_lua,
+		.lua_subrs = NULL,
 };
 
 /*
@@ -371,7 +372,7 @@ rspamadm_lua_message_handler (lua_State *L, gint argc, gchar **argv)
 			rspamd_printf ("cannot open %s: %s\n", argv[i], strerror (errno));
 		}
 		else {
-			task = rspamd_task_new (NULL, NULL, NULL);
+			task = rspamd_task_new (NULL, NULL, NULL, NULL);
 
 			if (!rspamd_task_load_message (task, NULL, map, len)) {
 				rspamd_printf ("cannot load %s\n", argv[i]);
@@ -548,7 +549,7 @@ rspamadm_lua_accept_cb (gint fd, short what, void *arg)
 		return;
 	}
 
-	session = g_slice_alloc0 (sizeof (*session));
+	session = g_malloc0 (sizeof (*session));
 	session->rt = ctx->rt;
 	session->ctx = ctx;
 	session->addr = addr;
@@ -571,7 +572,7 @@ rspamadm_lua_finish_handler (struct rspamd_http_connection_entry *conn_ent)
 {
 	struct rspamadm_lua_repl_session *session = conn_ent->ud;
 
-	g_slice_free1 (sizeof (*session), session);
+	g_free (session);
 }
 
 /*
@@ -740,7 +741,7 @@ rspamadm_lua (gint argc, gchar **argv)
 		}
 
 		ev_base = event_init ();
-		ctx = g_slice_alloc0  (sizeof (*ctx));
+		ctx = g_malloc0  (sizeof (*ctx));
 		http = rspamd_http_router_new (rspamadm_lua_error_handler,
 						rspamadm_lua_finish_handler,
 						NULL, ev_base,
@@ -758,7 +759,7 @@ rspamadm_lua (gint argc, gchar **argv)
 			if (fd != -1) {
 				struct event *ev;
 
-				ev = g_slice_alloc0 (sizeof (*ev));
+				ev = g_malloc0 (sizeof (*ev));
 				event_set (ev, fd, EV_READ|EV_PERSIST, rspamadm_lua_accept_cb,
 						ctx);
 				event_base_set (ev_base, ev);
