@@ -16,7 +16,7 @@ class MyTCPHandler(socketserver.BaseRequestHandler):
         os.remove(PID)
         self.data = self.request.recv(1024).strip()
         if self.server.foundvirus:
-            self.request.sendall(b"1 <infected: EICAR_Test_File> FOO\n")
+            self.request.sendall(b"1 <infected: EICAR_Test_File> FOO->bar\n")
         else:
             self.request.sendall(b"0 <clean> FOO\n")
         self.request.close()
@@ -32,7 +32,10 @@ if __name__ == "__main__":
     alen = len(sys.argv)
     if alen > 1:
         port = int(sys.argv[1])
-        if alen >= 3:
+        if alen >= 4:
+            PID = sys.argv[3]
+            foundvirus = bool(sys.argv[2])
+        elif alen >= 3:
             foundvirus = bool(sys.argv[2])
         else:
             foundvirus = False
